import 'dart:async';
import 'package:flutter/material.dart';
import 'package:tivio_sdk/player/tivio_player_input_source.dart';
import 'package:better_player/better_player.dart';
import 'package:better_player/src/video_player/video_player_platform_interface.dart'; // TD extract to own file
import 'better_player_material_progress_bar.dart';
import 'svg_button.dart';
import 'better_player_ui_press_controls.dart';
import 'better_player_error.dart';
import '../tivio_player_colors.dart';
import '../tivio_player_layout.dart';
import '../../enums/tivio_player_layout.dart';

class BetterPlayerControlUi extends StatelessWidget {
  final BetterPlayerUiPressControls controls;
  final TivioPlayerColors tivioPlayerColors;
  final TivioPlayerLayout tivioPlayerLayout;
  final void Function() resetHideUiAfterDelay;
  final void Function(Duration position) overwriteProgress;

  final bool isPlaying;
  final bool isInFullscreen;
  final bool isBuffering;

  final Duration? duration;
  final Duration? progress;

  final List<DurationRange> buffered;

  BetterPlayerControlUi({
    required this.controls,
    required this.tivioPlayerColors,
    required this.tivioPlayerLayout,
    required this.resetHideUiAfterDelay,
    required this.isBuffering,
    required this.overwriteProgress,
    required this.buffered,
    this.isPlaying = true,
    this.isInFullscreen = true,
    this.duration,
    this.progress,
  });

  String formatNumber(int number) {
    return (number > 9) ? number.toString() : '0$number';
  }

  String createCounterLabel() {
    int restOfSeconds = (duration != null && progress != null) ? duration!.inSeconds - progress!.inSeconds : 0;

    Duration restAsDuration = Duration(seconds: restOfSeconds);

    return '-${formatNumber(restAsDuration.inMinutes)}:${formatNumber(restAsDuration.inSeconds % 60)}';
  }

  @override
  Widget build(BuildContext context) {
    String counterLabel = createCounterLabel();

    return Padding(
      padding: EdgeInsets.all(10),
      child: Column(
        mainAxisAlignment: MainAxisAlignment.center,
        children: [
          // Custom top controls
          Container(
            height: 70,
            child: Row(
              mainAxisAlignment: MainAxisAlignment.end,
              crossAxisAlignment: CrossAxisAlignment.start,
              children: [
                (this.tivioPlayerLayout.topRight == TivioPlayerLayoutTopRight.custom &&
                        this.tivioPlayerLayout.topRightCustomWidget != null)
                    ? this.tivioPlayerLayout.topRightCustomWidget!
                    : Container(),
              ],
            ),
          ),
          // Middle controls
          Expanded(
            flex: 1,
            child: Container(
              child: Row(
                mainAxisAlignment: MainAxisAlignment.center,
                children: [
                  SvgButton(
                    iconName: 'backward',
                    onTap: () {
                      resetHideUiAfterDelay();
                      controls.onBackwardPressed();
                    },
                    width: 60,
                    height: 60,
                    padding: 5,
                  ),
                  SizedBox(width: 40),
                  (isBuffering)
                      ? Padding(
                          padding: EdgeInsets.only(left: 10, right: 10),
                          child: SizedBox(
                            child: CircularProgressIndicator(
                              color: Colors.white,
                              strokeWidth: 5,
                            ),
                            width: 50,
                            height: 50,
                          ),
                        )
                      : SvgButton(
                          iconName: (isPlaying) ? 'pause' : 'play',
                          onTap: () {
                            if (isPlaying) {
                              controls.onPausePressed();
                            } else {
                              controls.onPlayPressed();
                            }
                          },
                          width: 60,
                          height: 60,
                          padding: 5,
                        ),
                  SizedBox(width: 40),
                  SvgButton(
                    iconName: 'forward',
                    onTap: () {
                      resetHideUiAfterDelay();
                      controls.onForwardPressed();
                    },
                    width: 60,
                    height: 60,
                    padding: 5,
                  ),
                ],
              ),
            ),
          ),
          // Progress & full screen
          Container(
            height: 71,
            child: Column(
              mainAxisAlignment: MainAxisAlignment.end,
              children: [
                Row(
                  mainAxisAlignment: MainAxisAlignment.end,
                  children: [
                    (this.tivioPlayerLayout.bottomRight == TivioPlayerLayoutBottomRight.toggleFullscreen)
                        ? SvgButton(
                            iconName: (isInFullscreen) ? 'fullscreen-exit' : 'fullscreen-enter',
                            onTap: () {
                              resetHideUiAfterDelay();
                              if (isInFullscreen) {
                                controls.onExitFullscreenPressed();
                              } else {
                                controls.onEnterFullscreenPressed();
                              }
                            },
                            width: 40,
                          )
                        : Container(),
                  ],
                ),
                Row(
                  children: [
                    Expanded(
                      child: BetterPlayerMaterialVideoProgressBar(
                        colors: BetterPlayerProgressColors(
                          playedColor: tivioPlayerColors.progressBarPlayed,
                          bufferedColor: Color.fromRGBO(255, 255, 255, 0.2),
                          backgroundColor: Color.fromRGBO(255, 255, 255, 0.2),
                        ),
                        onSeekingEnded: (Duration position) {
                          resetHideUiAfterDelay();
                          overwriteProgress(position);
                          controls.onSeekingEnded(position);
                        },
                        progress: this.progress!,
                        duration: this.duration!,
                        buffered: this.buffered,
                        isPlaying: this.isPlaying,
                        onDragStart: () {
                          if (this.isPlaying) {
                            controls.onPausePressed();
                          }
                        },
                        onDragEnd: (bool wasPlaying) {
                          if (wasPlaying) {
                            controls.onPlayPressed();
                          }
                        },
                      ),
                    ),
                    Padding(
                      padding: EdgeInsets.only(left: 10, top: 10),
                      child: Text(
                        counterLabel,
                        style: TextStyle(
                          fontSize: 18,
                          color: Color.fromRGBO(255, 255, 255, 0.7),
                        ),
                      ),
                    ),
                  ],
                ),
              ],
            ),
          ),
        ],
      ),
    );
  }
}

class BetterPlayerControl extends StatefulWidget {
  final BetterPlayerUiPressControls controls;
  final BetterPlayerController controller;
  final TivioPlayerColors tivioPlayerColors;
  final TivioPlayerLayout tivioPlayerLayout;

  StreamController<BetterPlayerEvent> eventPlayerStream;
  TivioPlayerVodSource source;

  BetterPlayerControl({
    required this.controls,
    required this.controller,
    required this.tivioPlayerColors,
    required this.tivioPlayerLayout,
    required this.eventPlayerStream,
    required this.source,
  });

  @override
  createState() => _BetterPlayerControlState();
}

class _BetterPlayerControlState extends State<BetterPlayerControl> with SingleTickerProviderStateMixin {
  Duration? progress = Duration();
  StreamSubscription? subscription;
  bool isPlaying = true;
  bool isInFullscreen = true;
  bool isBuffering = false;
  bool hasError = false;
  Timer? showUiTimer;
  late AnimationController controller;

  @override
  void initState() {
    super.initState();

    if (subscription != null) {
      subscription!.cancel();
    }

    subscription = widget.eventPlayerStream.stream.listen((event) {
      checkForError();

      switch (event.betterPlayerEventType) {
        case BetterPlayerEventType.play:
          hideUiAfterDelay();
          setState(() {
            isPlaying = true;

            /// Fix buffering indicator is sometimes shown on iOS, even when video already playing
            isBuffering = false;
          });
          break;
        case BetterPlayerEventType.pause:
          showUiAndCancelDelay();
          setState(() {
            isPlaying = false;
          });
          break;
        case BetterPlayerEventType.progress:

          /// Fix do not change progress while buffering because iOS emits "random" values
          if (!isBuffering) {
            setState(() {
              progress = event.parameters!['progress'];
            });
          }
          break;
        case BetterPlayerEventType.finished:

          /// Fix missing last tick
          setState(() {
            progress = widget.controller.videoPlayerController!.value.duration;
          });
          break;
        case BetterPlayerEventType.hideFullscreen:
          setState(() {
            isInFullscreen = false;
          });
          break;
        case BetterPlayerEventType.openFullscreen:
          setState(() {
            isInFullscreen = true;
          });
          break;
        case BetterPlayerEventType.bufferingStart:
          showUiAndCancelDelay();
          setState(() {
            isBuffering = true;
          });
          break;
        case BetterPlayerEventType.bufferingEnd:
          hideUiAfterDelay();
          setState(() {
            isBuffering = false;
          });
          break;
        default:
          break;
      }
    });

    hideUiAfterDelay();
    setupAnimation();
  }

  checkForError() {
    setState(() {
      hasError = widget.controller.videoPlayerController?.value.hasError ?? false;
    });
  }

  showUi() {
    controller.animateTo(1);
  }

  hideUi() {
    controller.animateTo(0);
  }

  hideUiAfterDelay() {
    if (showUiTimer != null) {
      showUiTimer!.cancel();
    }
    showUiTimer = Timer(Duration(seconds: 3), () {
      hideUi();
    });
  }

  showUiAndCancelDelay() {
    if (showUiTimer != null) {
      showUiTimer!.cancel();
    }
    showUi();
  }

  showUiAndHideAfterDelay() {
    if (isPlaying) {
      showUi();
      hideUiAfterDelay();
    }
  }

  setupAnimation() {
    controller = AnimationController(
      value: 1,
      duration: Duration(milliseconds: 300),
      vsync: this,
    );
  }

  @override
  Widget build(BuildContext context) {
    return (hasError)
        ? BetterPlayerError(
            onPressed: () => widget.controller.retryDataSource(),
          )
        : InkWell(
            onTap: () => showUiAndHideAfterDelay(),
            child: FadeTransition(
              opacity: controller,
              child: Container(
                width: 100.0,
                height: 100.0,
                child: (widget.controller.videoPlayerController?.value != null &&
                        widget.controller.videoPlayerController!.value.initialized)
                    ? BetterPlayerControlUi(
                        controls: widget.controls,
                        tivioPlayerColors: widget.tivioPlayerColors,
                        tivioPlayerLayout: widget.tivioPlayerLayout,
                        isPlaying: isPlaying,
                        progress: progress,
                        duration: widget.controller.videoPlayerController!.value.duration,
                        isInFullscreen: isInFullscreen,
                        isBuffering: isBuffering,
                        resetHideUiAfterDelay: () => showUiAndHideAfterDelay(),
                        overwriteProgress: (Duration overwrittenProgress) {
                          setState(() {
                            progress = overwrittenProgress;
                          });
                        },
                        buffered: widget.controller.videoPlayerController!.value.buffered,
                      )
                    : null,
              ),
            ),
          );
  }

  @override
  void dispose() {
    if (subscription != null) {
      subscription!.cancel();
    }
    if (showUiTimer != null) {
      showUiTimer!.cancel();
    }
    controller.dispose();
    super.dispose();
  }
}
